#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <getopt.h>
#include <sys/stat.h>
#ifdef _WIN32
# include <io.h>
# include <fcntl.h>
#endif
#include <signal.h>
#include <sys/time.h>
#include "zbar.h"


// The following CRC code comes from RFC1952, Section 8
//   http://tools.ietf.org/html/rfc1952#section-8

/* Table of CRCs of all 8-bit messages. */
unsigned long crc_table[256];

/* Flag: has the table been computed? Initially false. */
int crc_table_computed = 0;

/* Make the table for a fast CRC. */
void make_crc_table(void)
{
  unsigned long c;
  int n, k;
  for (n = 0; n < 256; n++) {
    c = (unsigned long) n;
    for (k = 0; k < 8; k++) {
      if (c & 1) {
        c = 0xedb88320L ^ (c >> 1);
      } else {
        c = c >> 1;
      }
    }
    crc_table[n] = c;
  }
  crc_table_computed = 1;
}


/*
    Update a running crc with the bytes buf[0..len-1] and return
  the updated crc. The crc should be initialized to zero. Pre- and
  post-conditioning (one's complement) is performed within this
  function so it shouldn't be done by the caller. Usage example:

  unsigned long crc = 0L;

  while (read_buffer(buffer, length) != EOF) {
    crc = update_crc(crc, buffer, length);
  }
  if (crc != original_crc) error();
*/
static unsigned long update_crc(unsigned long crc,
  unsigned char *buf, int len)
{
  unsigned long c = crc ^ 0xffffffffL;
  int n;

  if (!crc_table_computed)
    make_crc_table();
  for (n = 0; n < len; n++) {
    c = crc_table[(c ^ buf[n]) & 0xff] ^ (c >> 8);
  }
  return c ^ 0xffffffffL;
}

/* Return the CRC of the bytes buf[0..len-1]. */
static unsigned long crc(unsigned char *buf, int len)
{
  return update_crc(0L, buf, len);
}




/*

                                   _JNJ`
                               .JNMH`
                              JMMF`       `;.
                            .NMM)          `MN.
                            MMM)            (MML
                           (MMM`             MMML
                     M  I  D  N  I  G  H  T    C  o  D  E
                           (NMMF             MHNH
                            NMML            .MMM
                             NMML          .NMH
                              4MMNL       .#F
                               `4HNNL_   `
                                   `"""`

                          Copyright (C) 2004-2014
       "Ian (Larry) Latter" <ian dot latter at midnightcode dot org>

    Midnight Code is a registered trademark of Ian Latter.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, and mirrored at the Midnight Code web
    site; as at version 2 of the License only.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    (version 2) along with this program; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
    02111-1307 USA, or see http://midnightcode.org/gplv2.txt

*/

#include "tgxf-receive.h"

#define TGXF_DEBUG 1

#define TGXF_CONTROL_TYPE_START 1
#define TGXF_CONTROL_TYPE_STOP 2
#define TGXF_CONTROL_TYPE_STATUS 3
#define TGXF_CONTROL_SUBTYPE_START_FILENAME 1
#define TGXF_CONTROL_SUBTYPE_START_FILESIZE 2
#define TGXF_CONTROL_SUBTYPE_START_QRCODE_VERSION 3
#define TGXF_CONTROL_SUBTYPE_START_QRCODE_FPS 4
#define TGXF_CONTROL_SUBTYPE_START_QRCODE_BYTES 5
#define TGXF_CONTROL_SUBTYPE_STOP_PAUSE 1
#define TGXF_CONTROL_SUBTYPE_STOP_COMPLETE 2
#define TGXF_CONTROL_SUBTYPE_STOP_CANCEL 3
#define TGXF_CONTROL_SUBTYPE_STATUS_SINCE 1

// TGXf Session Parameters (TGXf State Machine)
static unsigned int tgxf_session_total_frames = 0;
static unsigned int tgxf_session_first_data_frame = 0;
static unsigned int tgxf_session_good_start = 0;
static unsigned int tgxf_session_last_data_counter = 15;
static unsigned int tgxf_session_frame_count = 0;
static unsigned int tgxf_session_error_count = 0;
static unsigned long tgxf_session_recent_crc32 = 0L;
static unsigned long tgxf_session_total_crc32 = 0L;
static unsigned int tgxf_session_max_errors = 5;
static unsigned int tgxf_session_errors = 0;
static char * tgxf_session_error_str = NULL;
static unsigned int tgxf_session_error_size = 0;
static unsigned int tgxf_session_error_len = 0;
static unsigned int tgxf_session_aborted = 0;
static unsigned int tgxf_session_total_bytes = 0;
static unsigned int tgxf_session_complete = 0;
static unsigned int tgxf_session_max_timeout = 5;
static unsigned int tgxf_session_timeout_ticks = 0;
static unsigned int tgxf_session_paused = 0;

static unsigned int tgxf_session_debug_image_counter = 0;

// TGXf Session Parameters (TGXf Global Options)
static unsigned char tgxf_session_file_name[20];
static unsigned int tgxf_session_file_size = 0;
static unsigned int tgxf_session_qrcode_version = 0;
static unsigned int tgxf_session_qrcode_fps = 0;
static unsigned int tgxf_session_qrcode_bytes = 0;


static unsigned int get_size_for_file(char *filepath) {
  struct stat st;
  unsigned int size;

  stat(filepath, &st);
  size = st.st_size;

  return size;
}


static unsigned long get_crc_for_file(char *filepath) {
  unsigned char data_buffer[1024];
  unsigned char *data_buffer_p;
  int data_buffer_size = 1024;
  unsigned int read_len;
  unsigned long crc;
  FILE *fp;

  crc = 0L;
  if(!filepath)
    return crc;

  if((fp = fopen((const char *)filepath, "rb")) == NULL)
    return crc;

  data_buffer_p = data_buffer;
  while((read_len = fread(data_buffer, 1, data_buffer_size, fp)) > 0) {
    crc = update_crc(crc, data_buffer_p, read_len);
  }

  return crc;
}


static unsigned int unpack_control_payload_16bit_number(unsigned char *frame_payload, int payload_size)
{

  if(payload_size >= 2)
    return (unsigned int)frame_payload[0] << 8 | (unsigned int)frame_payload[1];

  return 0;
}


static unsigned long unpack_control_payload_32bit_number(unsigned char *frame_payload, int payload_size)
{
  if(payload_size >= 4) {
    return (unsigned long)frame_payload[0] << 24 | (unsigned long)frame_payload[1] << 16 |
      (unsigned long)frame_payload[2] << 8 | (unsigned long)frame_payload[3]; 
  }

  return 0L;
}


static unsigned int unpack_control_payload_clean_string(unsigned char *string_dest, unsigned char *frame_payload, int payload_size)
{
  unsigned char *chr_p;

  memset(string_dest, 0, 10);
  if(payload_size > 0) {
    // Store String
    if(memcpy(string_dest, frame_payload, 9) != string_dest) {
      fprintf(stderr, "ERROR(TGXf): badly formatted control string.\n");
      return -1;
    }
    string_dest[10] = '\0';
    // Clean String
    chr_p = string_dest;
    while(*chr_p) {
      // Unsavory characters are replaced
      if(*chr_p < 32 || ( *chr_p > 32 && *chr_p < 45 ) ||
        ( *chr_p > 46 && *chr_p < 48 ) || ( *chr_p > 57 && *chr_p < 65 ) ||
        ( *chr_p > 90 && *chr_p < 97 ) || *chr_p > 122) {
          *chr_p = '_';
      }
      // No leading dot
      if(*chr_p == 46 && chr_p == string_dest) {
          *chr_p = '_';
      }
      chr_p++;
    }
  }

  return 0;
}


void tgxf_timeout_watchdog(int sig) {
  if(tgxf_session_complete) {
    tgxf_session_timeout_ticks = 0;
    if(TGXF_DEBUG)
      printf("DEBUG(TGXf): Session complete ..\n");
    return;
  }
  if(tgxf_session_paused) {
    tgxf_session_timeout_ticks = 0;
    if(TGXF_DEBUG)
      printf("DEBUG(TGXf): Session paused ..\n");
    return;
  }
  if(tgxf_session_timeout_ticks >= tgxf_session_max_timeout) {
    if(!tgxf_session_aborted) {
      printf("(TGXf): Session receive timeout exceeded, aborting xfer.\n");
      tgxf_session_aborted = 1;
    }
  }
  tgxf_session_timeout_ticks++;

  return;
}

int tgxf_get_current_frame_num() {
    return tgxf_session_frame_count+1;
}
int tgxf_get_total_num_frames() {
    return tgxf_session_total_frames;
}
int tgxf_get_num_errors() {
    return tgxf_session_errors;
}
void tgxf_set_max_errors(int num) {
    tgxf_session_max_errors = num;
}

void tgxf_set_max_timeout(int num) {
    tgxf_session_max_timeout = num;
}

int tgxf_get_fps() {
    return tgxf_session_qrcode_fps;
}

const char * tgxf_get_filename() {
	return tgxf_session_file_name;
}

void tgxf_reset() {
    tgxf_session_total_frames = 0;
    tgxf_session_first_data_frame = 0;
    tgxf_session_good_start = 0;
    tgxf_session_last_data_counter = 15;
    tgxf_session_frame_count = 0;
    tgxf_session_error_count = 0;
    tgxf_session_recent_crc32 = 0L;
    tgxf_session_total_crc32 = 0L;
    tgxf_session_max_errors = 5;
    tgxf_session_errors = 0;
    tgxf_session_error_str = NULL;
    tgxf_session_error_size = 0;
    tgxf_session_error_len = 0;
    tgxf_session_aborted = 0;
    tgxf_session_total_bytes = 0;
    tgxf_session_complete = 0;
    tgxf_session_max_timeout = 5;
    tgxf_session_timeout_ticks = 0;
    tgxf_session_paused = 0;
    
    tgxf_session_debug_image_counter = 0;
    
    // TGXf Session Parameters (TGXf Global Options)
    tgxf_session_file_name[0] = 0;
    tgxf_session_file_size = 0;
    tgxf_session_qrcode_version = 0;
    tgxf_session_qrcode_fps = 0;
    tgxf_session_qrcode_bytes = 0;
}

enum TgxfReceiveState decode_tgxf_frame(const zbar_symbol_t *sym, const char * filePath) {
  struct itimerval tbuf;
  struct sigaction action;
  unsigned char *frame_buffer;
  unsigned char *frame_payload;
  char error_buffer[1024];
  int frame_size;
  int payload_size;
  int error_buffer_size = 1024;

  unsigned char control_byte;
  unsigned int control_bit;
  unsigned int control_type;
  unsigned int control_subtype;
  unsigned int data_counter;
  unsigned int expected_counter;
  unsigned int lost_frames;
  unsigned int transfer_bytes_remaining;
  unsigned int frames_remaining;
  unsigned int minutes_remaining;
  unsigned int seconds_remaining;
  unsigned int bytes_to_use;
  unsigned int lost_frame_idx;
  FILE *fp;
  int i;
  float progress_state;
    
  frame_size = zbar_symbol_get_data_length(sym);
  if(frame_size <= 0) {
    fprintf(stderr, "ERROR(TGXf): invalid frame size: %d\n", frame_size);
    return -1;
  }
  if((frame_buffer = (unsigned char *)malloc(frame_size)) == NULL) {
    fprintf(stderr, "ERROR(TGXf): unable to allocate frame buffer\n");
    return -1;
  }
  if(memcpy(frame_buffer, zbar_symbol_get_data(sym), frame_size) != frame_buffer) {
    fprintf(stderr, "ERROR(TGXf): failed to establish frame buffer\n");
    return -1;
  }

    // Write Data to File
    const char * filename = (const char *)tgxf_session_file_name;
    if (filePath) {
        filename = filePath;
    }
    
  // Update watchdog timer
  tgxf_session_timeout_ticks = 0;

  // Point to Payload
  frame_payload = frame_buffer;
  frame_payload++;
  payload_size = frame_size - 1;

  // Get Control Byte and Bit
  control_byte = *frame_buffer;
  control_bit = control_byte & 1;

  switch(control_bit) {

    // CONTROL FRAME
    case 1:
      if(TGXF_DEBUG)
        fprintf(stdout, "DEBUG(TGXf): Control Frame\n");
      control_type = (control_byte & 14) >> 1;
      // Only accept STOP if session has been aborted by client
      if(tgxf_session_aborted && control_type != TGXF_CONTROL_TYPE_STOP) {
        if(TGXF_DEBUG)
          fprintf(stdout, "DEBUG(TGXf): ignoring non-STOP control message in aborted transfer.\n");
        break;
      }
      control_subtype = (control_byte & 240) >> 4;
      if(TGXF_DEBUG)
        fprintf(stdout, "DEBUG(TGXf): Control Type=[%d], SubType=[%d]\n",control_type, control_subtype);
      switch(control_type) {
        case TGXF_CONTROL_TYPE_START:
          if(TGXF_DEBUG)
            fprintf(stdout, "DEBUG(TGXf): Control Type START\n");
          // Don't accept START if data frames have been received
          if(tgxf_session_first_data_frame) {
            if(TGXF_DEBUG)
              fprintf(stdout, "DEBUG(TGXf): ignoring late START control message in active transfer.\n");
          }
          switch(control_subtype) {
            case TGXF_CONTROL_SUBTYPE_START_FILENAME:
              if(TGXF_DEBUG)
                fprintf(stdout, "DEBUG(TGXf): Control SubType FILENAME\n");
                unpack_control_payload_clean_string(tgxf_session_file_name, frame_payload, payload_size);
              if(TGXF_DEBUG)
                fprintf(stdout, "DEBUG(TGXf): START/FILENAME [%s].\n", tgxf_session_file_name);
              break;
            case TGXF_CONTROL_SUBTYPE_START_FILESIZE:
              tgxf_session_file_size = unpack_control_payload_16bit_number(frame_payload, payload_size);
              if(TGXF_DEBUG)
                fprintf(stdout, "DEBUG(TGXf): START/FILESIZE [%d].\n", tgxf_session_file_size);
              break;
            case TGXF_CONTROL_SUBTYPE_START_QRCODE_VERSION:
              tgxf_session_qrcode_version = unpack_control_payload_16bit_number(frame_payload, payload_size);
              if(TGXF_DEBUG)
                fprintf(stdout, "DEBUG(TGXf): START/QRCODE_VERSION [%d].\n", tgxf_session_qrcode_version);
              break;
            case TGXF_CONTROL_SUBTYPE_START_QRCODE_FPS:
              tgxf_session_qrcode_fps = unpack_control_payload_16bit_number(frame_payload, payload_size);
              if(TGXF_DEBUG)
                fprintf(stdout, "DEBUG(TGXf): START/QRCODE_FPS [%d].\n", tgxf_session_qrcode_fps);
              break;
            case TGXF_CONTROL_SUBTYPE_START_QRCODE_BYTES:
              tgxf_session_qrcode_bytes = unpack_control_payload_16bit_number(frame_payload, payload_size);
              if(TGXF_DEBUG)
                fprintf(stdout, "DEBUG(TGXf): START/QRCODE_BYTES [%d].\n", tgxf_session_qrcode_bytes);
              break;
            default:
              fprintf(stdout, "WARNING(TGXf): reserved control sub-type used, frame ignored.\n");
              free(frame_buffer);
              return TGXF_RECEIVE_STATE_IN_PROGRESS;
          }
          break;

        case TGXF_CONTROL_TYPE_STOP:
          if(TGXF_DEBUG)
            fprintf(stdout, "DEBUG(TGXf): Control Type STOP\n");
          switch(control_subtype) {
            case TGXF_CONTROL_SUBTYPE_STOP_PAUSE:
              tgxf_session_paused = !tgxf_session_paused;
              if(TGXF_DEBUG)
                fprintf(stdout, "DEBUG(TGXf): STOP/PAUSE.\n");
              break;
            case TGXF_CONTROL_SUBTYPE_STOP_COMPLETE:
              tgxf_session_total_crc32 = unpack_control_payload_32bit_number(frame_payload, payload_size);
              if(!tgxf_session_good_start) {
                if(TGXF_DEBUG)
                  fprintf(stdout, "WARNING(TGXf): failed to get required start data, ignoring out of state STOP.\n");
                break;
              }
              tgxf_session_recent_crc32 = get_crc_for_file((char *)filename);
              if(tgxf_session_total_crc32 != tgxf_session_recent_crc32) {
                fprintf(stdout, "ERROR(TGXf): CRC doesn't match, calculated %lu, received %lu.\n",
                  tgxf_session_recent_crc32, tgxf_session_total_crc32);
                   tgxf_session_complete = 1;// JAMES we may have a partial file even though CRC failed
              } else {
                tgxf_session_complete = 1;
                fprintf(stdout, "(TGXf): CRC validated, transfer successful.\n");
              }
              // if(TGXF_DEBUG)
                // fprintf(stdout, "DEBUG(TGXf): STOP/COMPLETE [%u].\n", tgxf_session_total_crc32);
              // SESSION RESET
              break;
            case TGXF_CONTROL_SUBTYPE_STOP_CANCEL:
              if(TGXF_DEBUG)
                fprintf(stdout, "DEBUG(TGXf): STOP/CANCEL.\n");
              // SESSION RESET
              break;
            default:
              fprintf(stdout, "WARNING(TGXf): reserved control sub-type used, frame ignored.\n");
              free(frame_buffer);
              return TGXF_RECEIVE_STATE_IN_PROGRESS;
          }
          break;

        case TGXF_CONTROL_TYPE_STATUS:
          if(TGXF_DEBUG)
            fprintf(stdout, "DEBUG(TGXf): Control Type STATUS\n");
          switch(control_subtype) {
            case TGXF_CONTROL_SUBTYPE_STATUS_SINCE:
              tgxf_session_recent_crc32 = unpack_control_payload_32bit_number(frame_payload, payload_size);
              // if(TGXF_DEBUG)
                // fprintf(stdout, "DEBUG(TGXf): STATUS/SINCE [%u].\n", tgxf_session_recent_crc32);
              break;
            default:
              fprintf(stdout, "WARNING(TGXf): reserved control sub-type used, frame ignored.\n");
              free(frame_buffer);
              return -1;
          }
          break;

        default:
          fprintf(stdout, "WARNING(TGXf): reserved control type used, frame ignored.\n");
          free(frame_buffer);
          return TGXF_RECEIVE_STATE_IN_PROGRESS;
      }
      break;

    // DATA FRAME
    case 0:
      if(TGXF_DEBUG)
        fprintf(stdout, "DEBUG(TGXf): Data Frame ------------------------------\n");

      // Ignore data frames if session aborted
      if(tgxf_session_aborted) {
        if(TGXF_DEBUG)
          fprintf(stdout, "WARNING(TGXf): session aborted, ignoring data frames.\n");
        break;
      }

      // If this is the first Data Frame then do we have the right START info to continue?
      if(!tgxf_session_first_data_frame) {
        tgxf_session_first_data_frame = 1;
        fprintf(stdout, "-= TGXf =-\n");
        fprintf(stdout, "  Filename:                 %s\n", tgxf_session_file_name);
        fprintf(stdout, "  File size:                %d\n", tgxf_session_file_size);
        fprintf(stdout, "  QRcode Version:           %d\n", tgxf_session_qrcode_version);
        fprintf(stdout, "  QRcode Frames per Second: %d\n", tgxf_session_qrcode_fps);
        fprintf(stdout, "  QRcode Bytes per Frame:   %d\n", tgxf_session_qrcode_bytes);
        fprintf(stdout, "\n");
        // Do we have sufficient START information to perform the data transfer?
        if(strlen((const char *)tgxf_session_file_name) > 0 &&
          tgxf_session_file_size > 0 &&
          tgxf_session_qrcode_fps > 0 &&
          tgxf_session_qrcode_bytes > 0) {

            // Establish session state
            tgxf_session_good_start = 1;

            // Setup session calculations
            tgxf_session_total_frames = tgxf_session_file_size / tgxf_session_qrcode_bytes;
            if(tgxf_session_total_frames * tgxf_session_qrcode_bytes < tgxf_session_file_size)
              tgxf_session_total_frames++;

            // Setup the TGXf watchdog timer
            action.sa_handler = tgxf_timeout_watchdog;
            sigemptyset(&action.sa_mask);
            action.sa_flags = 0;
            if(sigaction(SIGALRM, &action, NULL) < 0) {
              fprintf(stderr, "ERROR(TGXf): Watchdog timer setup failed, session won't timeout.\n");
            } else {
              tbuf.it_interval.tv_sec  = 1;
              tbuf.it_interval.tv_usec = 0;
              tbuf.it_value.tv_sec = 1;
              tbuf.it_value.tv_usec = 0;
              if(setitimer(ITIMER_REAL, &tbuf, NULL) == -1)
                fprintf(stderr, "ERROR(TGXf): Watchdog timer setup failed, session won't timeout.\n");
            }
        }
      }
      if(!tgxf_session_good_start) {
        if(TGXF_DEBUG)
          fprintf(stdout, "WARNING(TGXf): failed to get required start data, session aborted.\n");
        tgxf_session_aborted = 1;
        break;
      }

      // Duplicate Data Frame Detection - Ignore
      data_counter = (control_byte & 30) >> 1;
      if(data_counter == tgxf_session_last_data_counter) {
        if(TGXF_DEBUG)
          fprintf(stdout, "WARNING(TGXf): duplicate data frame, frame ignored.\n");
        break;
      }

      // Missing Data Frame Detection - Accumulate Errors
      expected_counter = tgxf_session_last_data_counter + 1;
      if(expected_counter == 16)
        expected_counter = 0;
      if(TGXF_DEBUG)
        fprintf(stdout, "DEBUG(TGXf): data_counter=%02d expected_counter=%02d; frame %d.\n",
          data_counter, expected_counter, tgxf_session_frame_count);
      if(data_counter != expected_counter) {
        lost_frames = data_counter - expected_counter;
        if(expected_counter > data_counter) {
          lost_frames = data_counter + 16 - expected_counter;
        }
        // Error Accumulation
        tgxf_session_errors += lost_frames;
        if(tgxf_session_errors >= tgxf_session_max_errors) {
          tgxf_session_aborted = 1;
          fprintf(stdout, "ERROR(TGXf): maximum errors for session reached, transfer aborted.\n");
        }
        // Accumulate lost frame numbers to report to user
        if(!tgxf_session_error_str) {
          tgxf_session_error_size = 1024;
          if((tgxf_session_error_str = (char *)malloc(tgxf_session_error_size)) == NULL) {
            fprintf(stderr, "ERROR(TGXf): unable to allocate error buffer\n");
            free(frame_buffer);
            return 1;
          }
          memset(tgxf_session_error_str, 0, tgxf_session_error_size);
        }
        for(lost_frame_idx = 0; lost_frame_idx < lost_frames; lost_frame_idx++) {
          memset(error_buffer, 0, error_buffer_size);
          if(strlen(tgxf_session_error_str) > 0) {
            snprintf(error_buffer, error_buffer_size, ",%d",
              tgxf_session_frame_count + lost_frame_idx + 1);
          } else {
            snprintf(error_buffer, error_buffer_size, "%d",
              tgxf_session_frame_count + lost_frame_idx + 1);
          }
          if(strlen(error_buffer) > tgxf_session_error_size - strlen(tgxf_session_error_str) - 2) {
            tgxf_session_error_size += 1024;
            if((tgxf_session_error_str =
              (char *)realloc(tgxf_session_error_str, tgxf_session_error_size)) == NULL) {
              fprintf(stderr, "ERROR(TGXf): unable to allocate error buffer\n");
              free(frame_buffer);
              return TGXF_RECEIVE_STATE_FAILED;
            }
          }
          strcat(tgxf_session_error_str, error_buffer);
          if(tgxf_session_aborted) {
            strcat(tgxf_session_error_str, "+");
            break;
          }
        }
        if(lost_frames == 1) {
          fprintf(stdout, "WARNING(TGXf): lost 1 data frame, error recorded.\n");
        } else {
          fprintf(stdout, "WARNING(TGXf): lost %d data frames, error recorded.\n", lost_frames);
        }
        // Correct the session frame count 
        tgxf_session_frame_count += lost_frames;
      }

      // Do last-packet calculation and under-run detection
      transfer_bytes_remaining = 
        tgxf_session_file_size - (tgxf_session_qrcode_bytes * tgxf_session_frame_count);
      bytes_to_use = tgxf_session_qrcode_bytes;
      if(transfer_bytes_remaining < tgxf_session_qrcode_bytes)
        bytes_to_use = transfer_bytes_remaining;
      tgxf_session_total_bytes += bytes_to_use;
      if(TGXF_DEBUG) {
        fprintf(stdout, "DEBUG(TGXf): payload_size=%d bytes_to_use=%d qrcode_bytes=%d; frame %d.\n", payload_size, bytes_to_use, tgxf_session_qrcode_bytes, tgxf_session_frame_count);
        fprintf(stdout, "DEBUG(TGXf): session_total_bytes after this frame is written = %d/%d.\n", tgxf_session_total_bytes, tgxf_session_file_size);
      }
      if(payload_size < bytes_to_use) {
        fprintf(stdout, "ERROR(TGXf): undersized payload on frame %d.\n", tgxf_session_frame_count);
        break;
      }

      // Update Display
      if(tgxf_session_errors) {
        if(tgxf_session_error_len < strlen(tgxf_session_error_str)) {
          tgxf_session_error_len = strlen(tgxf_session_error_str);
          fprintf(stdout, "\n");
          fprintf(stdout, "(TGXf): Frame errors: %s\n", tgxf_session_error_str);
        }
      }
      frames_remaining = tgxf_session_total_frames - tgxf_session_frame_count;
      seconds_remaining = frames_remaining / tgxf_session_qrcode_fps;
      minutes_remaining = seconds_remaining / 60;
      seconds_remaining = seconds_remaining % 60;
      progress_state = (float)20 - ((float)frames_remaining / (float)tgxf_session_total_frames * (float)20);
      fprintf(stdout, "(TGXf): Frame: %4d/%d ",
        tgxf_session_frame_count + 1, tgxf_session_total_frames);
      fprintf(stdout, " Errors: %2d/%d ",
        tgxf_session_errors, tgxf_session_max_errors);
      fprintf(stdout, " [");
      for(i = 0; i < 20; i++ ) {
        if(i >= progress_state) {
          fprintf(stdout, "%c", ' ' );
        } else {
          fprintf(stdout, "%c", '=' );
        }
      }
      fprintf(stdout, "] %02d min %02d sec.\r",
        minutes_remaining, seconds_remaining);
      if(TGXF_DEBUG || frames_remaining == 1)
        fprintf(stdout, "\n");
          
      if(access(filename, F_OK) == -1) {
        if((fp = fopen(filename, "w+b")) == NULL) {
          fprintf(stdout, "ERROR(TGXf): unable to create/open file for reading/writing, at %s.\n", filename);
          break;
        }
      } else {
        if((fp = fopen(filename, "r+b")) == NULL) {
          fprintf(stdout, "ERROR(TGXf): unable to open file for reading/writing, at %s.\n", filename);
          break;
        }
        if(fseek(fp, (tgxf_session_qrcode_bytes * tgxf_session_frame_count), SEEK_SET) < 0) {
          fprintf(stdout, "ERROR(TGXf): unable to seek file, at %s.\n", tgxf_session_file_name);
          fclose(fp);
          break;
        }
      }
      if(fwrite((const void *)frame_payload, (size_t)bytes_to_use, (size_t)1, fp) != (size_t)1) {
        fprintf(stdout, "ERROR(TGXf): unable to write to file, at %s.\n", filename);
        fclose(fp);
        break;
      }
      fclose(fp);

      // Increment tally
      tgxf_session_last_data_counter = data_counter;
      tgxf_session_frame_count++;
      break;

  }

  fflush(stdout);
  free(frame_buffer);
    
    if (tgxf_session_aborted) {
        return TGXF_RECEIVE_STATE_FAILED;
    }
    else if (tgxf_session_complete) {
        return TGXF_RECEIVE_STATE_SUCCEEDED;
    }
    return TGXF_RECEIVE_STATE_IN_PROGRESS;
}


static void barcode_handler(zbar_image_t *img, const void *userdata) {
  const zbar_symbol_t *barsym;
  zbar_symbol_type_t type;

  barsym = zbar_image_first_symbol(img);
  for(; barsym; barsym = zbar_symbol_next(barsym)) {
    if(zbar_symbol_get_count(barsym))
      continue;
    if((type = zbar_symbol_get_type(barsym)) == ZBAR_PARTIAL)
      continue;
    decode_tgxf_frame(barsym, NULL);
  }
}


